## Modeling Gas Flow Directions as State Variables:
# Does It Provide More Flexibility to Power Systems?
# Power Systems Computation Conference(PSCC) 2022, Electric Power Systems Research
# Code written by Junesoo Shin, s184386@student.dtu.dk
# 23.04.2022

function LP(data,ind,linepack,hours)
	@info("start model input")
	### Model
	LP_model = Model(Gurobi.Optimizer)
	
	### Variables
	@variable(LP_model, p[i in ind[:I], t in ind[:T]] >= 0) #PP dispatch
	@variable(LP_model, w[j in ind[:J], t in ind[:T]] >= 0) #Windfarm dispatch
	@variable(LP_model, -pi <= theta[n in ind[:N], t in ind[:T]] <= pi) #Voltage angle at node n
	@variable(LP_model, f[l in ind[:L], t in ind[:T]]) #Flow in transmission
	@variable(LP_model, g[k in ind[:K], t in ind[:T]] >= 0) #Gas dispatch
	@variable(LP_model, pr[mz in ind[:MZ], t in ind[:T]]) #Pressure at NGnodes
	@variable(LP_model, q_in_plus[z in ind[:Z], t in ind[:T]] >= 0) #Inflow natural gas rate of pipeline
	@variable(LP_model, q_out_plus[z in ind[:Z], t in ind[:T]] >= 0) #Outflow natural gas rate of pipeline
	@variable(LP_model, q[z in ind[:Z], t in ind[:T]]) #Natural gas flow in pipeline
	if linepack == true
		@variable(LP_model, h[z in ind[:Z], t in ind[:T]] >= 0) #Linepack in pipeline
	end

	### Objective function -> minimize cost of running TPPs and NGFPPs
	@objective(LP_model, Min,
			  sum(p[i,t]*data[:PP][i]["C_E"] for i in ind[:C], t in ind[:T]) #TPP costs
			  +sum(g[k,t]*data[:GasSupply][k]["C_G"] for k in ind[:K], t in ind[:T])) #NGFPPs costs

	### Power system constraints
	@constraint(LP_model, [i in ind[:I], t in ind[:T]], p[i,t] <= data[:PP][i]["Pmax"]) #PP supply limit
	@constraint(LP_model, [j in ind[:J], t in ind[:T]], w[j,t] <= data[:Windfarm][j]["Wmax"]*data[:WindFactor][t][j]) #WF supply forecast limit
	#Flow constraints (flow is increased x100 so that theta can take values between -pi and pi)
	@constraint(LP_model, [l in ind[:L], t in ind[:T]], f[l,t] == 100*data[:Lines][l]["B"]*(theta[data[:Lines][l]["From"],t]-theta[data[:Lines][l]["To"],t]))
	@constraint(LP_model, [l in ind[:L], t in ind[:T]], -data[:Lines][l]["capacity"] <= f[l,t] <= data[:Lines][l]["capacity"])
	@constraint(LP_model, [t in ind[:T]], theta["n1",t] == 0) #Reference bus
	#Nodal power balance
	@constraint(LP_model, power_bal[n in ind[:N], t in ind[:T]],
	sum(p[i,t] for i in ind[:I] if n==data[:PP][i]["ELnode"])
	+sum(w[j,t] for j in ind[:J] if n==data[:Windfarm][j]["ELnode"])
	-sum(f[l,t] for l in ind[:L] if n==data[:Lines][l]["From"])
	+sum(f[l,t] for l in ind[:L] if n==data[:Lines][l]["To"])
	== sum(data[:ELDemand][d]["share"]*data[:Demand][t]["EL"] for d in ind[:ELD] if data[:ELDemand][d]["node"]==n))

	### Gas system constraints
	#m_node = data[:Pipelines][z]["From"]
	#u_node = data[:Pipelines][z]["To"]
	@constraint(LP_model, [k in ind[:K], t in ind[:T]], g[k,t] <= data[:GasSupply][k]["Gmax"]) #Gas supply limit
	@constraint(LP_model, [mz in ind[:MZ], t in ind[:T]], data[:NGPr][mz]["minPr"] <= pr[mz,t] <= data[:NGPr][mz]["maxPr"]) #Pressure limit
	#Compressor factor
	@constraint(LP_model, [z in ind[:Z], t in ind[:T]; data[:Pipelines][z]["Gamma"]!=1], pr[data[:Pipelines][z]["To"],t] <= data[:Pipelines][z]["Gamma"]*pr[data[:Pipelines][z]["From"],t])

	#Linear approximation of Weymouth equation
	@constraint(LP_model, [z in ind[:Z], v in ind[:V], t in ind[:T];data[:KI_plus][z][v]>0],
	q[z,t] <= data[:KI_plus][z][v]*pr[data[:Pipelines][z]["From"],t]
			  -data[:KO_plus][z][v]*pr[data[:Pipelines][z]["To"],t])
	#More accurate operational constraints
	@constraint(LP_model, [z in ind[:Z], t in ind[:T]], pr[data[:Pipelines][z]["From"],t] >= pr[data[:Pipelines][z]["To"],t])
	@constraint(LP_model, [z in ind[:Z], t in ind[:T]], q[z,t] <= ind[:BigM]*(pr[data[:Pipelines][z]["From"],t] - pr[data[:Pipelines][z]["To"],t]))
	#Average flow
	@constraint(LP_model, [z in ind[:Z], t in ind[:T]], q[z,t] == 0.5*(q_in_plus[z,t]+q_out_plus[z,t]))

	if linepack == true
		#Linepack average mass
		@constraint(LP_model, [z in ind[:Z], t in ind[:T]], h[z,t] == data[:Pipelines][z]["K_h"]*0.5*(pr[data[:Pipelines][z]["From"],t] + pr[data[:Pipelines][z]["To"],t]))
		#Linepack mass conservation with initial and final(minimum) amount of linepack
		@constraint(LP_model, h_tstart[z in ind[:Z], t in ind[:T]; t=="t1"], h[z,t] == data[:Pipelines][z]["H_ini"] + q_in_plus[z,t] - q_out_plus[z,t])
		@constraint(LP_model, h_trest[z in ind[:Z], t in ind[:T]; t!="t1"], h[z,t] == h[z,data[:Demand][t]["prev"]] + q_in_plus[z,t] - q_out_plus[z,t])
		@constraint(LP_model, h_tend[z in ind[:Z], t in ind[:T]; t=="t$(hours)"], data[:Pipelines][z]["H_ini"] <= h[z,t])
	elseif linepack ==  false
		#To make sure what is sent is received in the pipelines (no linepack)
		@constraint(LP_model, [z in ind[:Z], t in ind[:T]], q_in_plus[z,t] == q_out_plus[z,t])
	end

	#Nodal gas balance
	@constraint(LP_model, gas_bal[mz in ind[:MZ], t in ind[:T]],
	sum(g[k,t] for k in ind[:K] if data[:GasSupply][k]["NGnode"] == mz)
	-sum(data[:PP][i]["eta"]*p[i,t] for i in ind[:G] if data[:PP][i]["NGnode"] == mz)
	-sum(q_in_plus[z,t] for z in ind[:Z] if data[:Pipelines][z]["From"] == mz)
	+sum(q_out_plus[z,t] for z in ind[:Z] if data[:Pipelines][z]["To"] == mz)
	==sum(data[:NGDemand][d]["share"]*data[:Demand][t]["NG"] for d in ind[:NGD] if data[:NGDemand][d]["node"] == mz))

	#Solve
	@info("start solving model")
	optimize!(LP_model)
	@info("LP model terminates with status: $(termination_status(LP_model))")
	if termination_status(LP_model) == MOI.OPTIMAL
		solution = Dict(
		:obj => objective_value(LP_model),
		:comp_time =>MOI.get(LP_model, MOI.SolveTime()),
		:p => [value.(p[i,t]) for i in ind[:I], t in ind[:T]],
		:w => [value.(w[j,t]) for j in ind[:J], t in ind[:T]],
		:theta => [value.(theta[n,t]) for n in ind[:N], t in ind[:T]],
		:f => [value.(f[l,t]) for l in ind[:L], t in ind[:T]],
		:g => [value.(g[k,t]) for k in ind[:K], t in ind[:T]],
		:pr => [value.(pr[mz,t]) for mz in ind[:MZ], t in ind[:T]],
		:q_in_plus => [value.(q_in_plus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q_out_plus => [value.(q_out_plus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q => [value.(q[z,t]) for z in ind[:Z], t in ind[:T]],
		:h => linepack == true ? [value.(h[z,t]) for z in ind[:Z], t in ind[:T]] : NaN)
		@info("New LP solution saved!")
		return solution
	else
		@info("No optimal solution available")
	end
end
