## Modeling Gas Flow Directions as State Variables:
# Does It Provide More Flexibility to Power Systems?
# Power Systems Computation Conference(PSCC) 2022, Electric Power Systems Research
# Code written by Junesoo Shin, s184386@student.dtu.dk
# 23.04.2022

function MILP(data,ind,linepack,mipgap,hours)
	@info("start model input")
	### Model
	MILP_model = Model(Gurobi.Optimizer)
	#set_optimizer_attribute(MILP_model, "MIPGap", mipgap) #Disable to allow default gap

	### Variables
	@variable(MILP_model, p[i in ind[:I], t in ind[:T]] >= 0) #PP dispatch
	@variable(MILP_model, w[j in ind[:J], t in ind[:T]] >= 0) #Windfarm dispatch
	@variable(MILP_model, -pi <= theta[n in ind[:N], t in ind[:T]] <= pi) #Voltage angle at node n
	@variable(MILP_model, f[l in ind[:L], t in ind[:T]]) #Flow in transmission
	@variable(MILP_model, g[k in ind[:K], t in ind[:T]] >= 0) #Gas dispatch
	@variable(MILP_model, pr[mz in ind[:MZ], t in ind[:T]]) #Pressure at NGnodes
	@variable(MILP_model, q_in_plus[z in ind[:Z], t in ind[:T]] >= 0) #Inflow natural gas rate of pipeline
	@variable(MILP_model, q_in_minus[z in ind[:Z], t in ind[:T]] >= 0) #Inflow natural gas rate of pipeline
	@variable(MILP_model, q_out_plus[z in ind[:Z], t in ind[:T]] >= 0) #Outflow natural gas rate of pipeline
	@variable(MILP_model, q_out_minus[z in ind[:Z], t in ind[:T]] >= 0) #Outflow natural gas rate of pipeline
	@variable(MILP_model, q_plus[z in ind[:Z], t in ind[:T]] >= 0) #Natural gas flow in pipeline from m to u
	@variable(MILP_model, q_minus[z in ind[:Z], t in ind[:T]] >= 0) #Natural gas flow in pipeline from m to u
	@variable(MILP_model, q[z in ind[:Z], t in ind[:T]]) #Natural gas flow in pipeline
	@variable(MILP_model, y[z in ind[:Z], t in ind[:T]], Bin) #Binary variable for defining direction of the flow of NG in pipeline
	if linepack == true
		@variable(MILP_model, h[z in ind[:Z], t in ind[:T]] >= 0) #Linepack in pipeline
	end
	@variable(MILP_model, phi_m[z in ind[:Z], t in ind[:T]] >= 0) #Auxiliary variable for binary expansion
	@variable(MILP_model, phi_u[z in ind[:Z], t in ind[:T]] >= 0) #Auxiliary variable for binary expansion

	### Objective function -> minimize cost of running TPPs and NGFPPs
	@objective(MILP_model, Min,
			  sum(p[i,t]*data[:PP][i]["C_E"] for i in ind[:C], t in ind[:T]) #TPP costs
			  +sum(g[k,t]*data[:GasSupply][k]["C_G"] for k in ind[:K], t in ind[:T])) #NGFPPs costs

	### Power system constraints
	@constraint(MILP_model, [i in ind[:I], t in ind[:T]], p[i,t] <= data[:PP][i]["Pmax"]) #PP supply limit
	@constraint(MILP_model, [j in ind[:J], t in ind[:T]], w[j,t] <= data[:Windfarm][j]["Wmax"]*data[:WindFactor][t][j]) #WF supply forecast limit
	#Flow constraints (flow is increased x100 so that theta can take values between -pi and pi)
	@constraint(MILP_model, [l in ind[:L], t in ind[:T]], f[l,t] == 100*data[:Lines][l]["B"]*(theta[data[:Lines][l]["From"],t]-theta[data[:Lines][l]["To"],t]))
	@constraint(MILP_model, [l in ind[:L], t in ind[:T]], -data[:Lines][l]["capacity"] <= f[l,t] <= data[:Lines][l]["capacity"])
	@constraint(MILP_model, [t in ind[:T]], theta["n1",t] == 0) #Reference bus
	#Nodal power balance
	@constraint(MILP_model, power_bal[n in ind[:N], t in ind[:T]],
	sum(p[i,t] for i in ind[:I] if n==data[:PP][i]["ELnode"])
	+sum(w[j,t] for j in ind[:J] if n==data[:Windfarm][j]["ELnode"])
	-sum(f[l,t] for l in ind[:L] if n==data[:Lines][l]["From"])
	+sum(f[l,t] for l in ind[:L] if n==data[:Lines][l]["To"])
	== sum(data[:ELDemand][d]["share"]*data[:Demand][t]["EL"] for d in ind[:ELD] if data[:ELDemand][d]["node"]==n))

	### Gas system constraints
	#m_node = data[:Pipelines][z]["From"]
	#u_node = data[:Pipelines][z]["To"]
	@constraint(MILP_model, [k in ind[:K], t in ind[:T]], g[k,t] <= data[:GasSupply][k]["Gmax"]) #Gas supply limit
	@constraint(MILP_model, [mz in ind[:MZ], t in ind[:T]], data[:NGPr][mz]["minPr"] <= pr[mz,t] <= data[:NGPr][mz]["maxPr"]) #Pressure limit
	#Compressor factor
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]; data[:Pipelines][z]["Gamma"]!=1], pr[data[:Pipelines][z]["To"],t] <= data[:Pipelines][z]["Gamma"]*pr[data[:Pipelines][z]["From"],t])

	#Linear approximation of Weymouth equation
	@constraint(MILP_model, flow_plus[z in ind[:Z], v in ind[:V], t in ind[:T];data[:KI_plus][z][v]>0],
	q_plus[z,t] <= data[:KI_plus][z][v]*pr[data[:Pipelines][z]["From"],t]
				   -data[:KO_plus][z][v]*pr[data[:Pipelines][z]["To"],t]
				   + ind[:BigM]*(1-y[z,t]))
	@constraint(MILP_model, flow_minus[z in ind[:Z], v in ind[:V], t in ind[:T];data[:KI_minus][z][v]>0],
	q_minus[z,t] <= data[:KI_minus][z][v]*pr[data[:Pipelines][z]["To"],t]
				   -data[:KO_minus][z][v]*pr[data[:Pipelines][z]["From"],t]
				   + ind[:BigM]*y[z,t])

	#Average flow
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_plus[z,t] == 0.5*(q_in_plus[z,t]+q_out_plus[z,t]))
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_minus[z,t] == 0.5*(q_in_minus[z,t]+q_out_minus[z,t]))
	#Bidirectional flow
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q[z,t] == q_plus[z,t]-q_minus[z,t])
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_plus[z,t] <= ind[:BigM]*(y[z,t]))
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_minus[z,t] <= ind[:BigM]*(1-y[z,t]))

	#Binary expansion
	@constraint(MILP_model, phi_def1[z in ind[:Z], t in ind[:T]], phi_m[z,t]-phi_u[z,t] >= 0)
	@constraint(MILP_model, phi_def2[z in ind[:Z], t in ind[:T]], pr[data[:Pipelines][z]["To"],t] - pr[data[:Pipelines][z]["From"],t] - phi_u[z,t] + phi_m[z,t] >= 0)

	@constraint(MILP_model, phi_m_def1[z in ind[:Z], t in ind[:T]], -y[z,t]*ind[:BigM] <= phi_m[z,t])
	@constraint(MILP_model, phi_m_def2[z in ind[:Z], t in ind[:T]], phi_m[z,t] <= y[z,t]*ind[:BigM])
	@constraint(MILP_model, phi_m_def3[z in ind[:Z], t in ind[:T]], -(1-y[z,t])*ind[:BigM] <= phi_m[z,t]-pr[data[:Pipelines][z]["From"],t])
	@constraint(MILP_model, phi_m_def4[z in ind[:Z], t in ind[:T]], phi_m[z,t]-pr[data[:Pipelines][z]["From"],t] <= (1-y[z,t])*ind[:BigM])

	@constraint(MILP_model, phi_u_def1[z in ind[:Z], t in ind[:T]], -y[z,t]*ind[:BigM] <= phi_u[z,t])
	@constraint(MILP_model, phi_u_def2[z in ind[:Z], t in ind[:T]], phi_u[z,t] <= y[z,t]*ind[:BigM])
	@constraint(MILP_model, phi_u_def3[z in ind[:Z], t in ind[:T]], -(1-y[z,t])*ind[:BigM] <= phi_u[z,t]-pr[data[:Pipelines][z]["To"],t])
	@constraint(MILP_model, phi_u_def4[z in ind[:Z], t in ind[:T]], phi_u[z,t]-pr[data[:Pipelines][z]["To"],t] <= (1-y[z,t])*ind[:BigM])

	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_plus[z,t] <= ind[:BigM]*(phi_m[z,t]-phi_u[z,t]))
	@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_minus[z,t] <= ind[:BigM]*(pr[data[:Pipelines][z]["To"],t] - pr[data[:Pipelines][z]["From"],t] - phi_u[z,t] + phi_m[z,t]))

	if linepack == true
		#Linepack average mass
		@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], h[z,t] == data[:Pipelines][z]["K_h"]*0.5*(pr[data[:Pipelines][z]["From"],t] + pr[data[:Pipelines][z]["To"],t]))
		#Linepack mass conservation with initial and final(minimum) amount of linepack
		@constraint(MILP_model, [z in ind[:Z], t in ind[:T]; t=="t1"], h[z,t] == data[:Pipelines][z]["H_ini"] + q_in_plus[z,t] - q_out_plus[z,t] + q_in_minus[z,t] - q_out_minus[z,t])
		@constraint(MILP_model, [z in ind[:Z], t in ind[:T]; t!="t1"], h[z,t] == h[z,data[:Demand][t]["prev"]] + q_in_plus[z,t] - q_out_plus[z,t] + q_in_minus[z,t] - q_out_minus[z,t])
		@constraint(MILP_model, [z in ind[:Z], t in ind[:T]; t=="t$(hours)"], data[:Pipelines][z]["H_ini"] <= h[z,t])
	elseif linepack ==  false
		#To make sure what is sent is received in the pipelines (no linepack)
		@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_in_plus[z,t] == q_out_plus[z,t])
		@constraint(MILP_model, [z in ind[:Z], t in ind[:T]], q_in_minus[z,t] == q_out_minus[z,t])
	end

	#Nodal gas balance
	@constraint(MILP_model, power_gas_bal[mz in ind[:MZ], t in ind[:T]],
	sum(g[k,t] for k in ind[:K] if data[:GasSupply][k]["NGnode"] == mz)
	-sum(data[:PP][i]["eta"]*p[i,t] for i in ind[:G] if data[:PP][i]["NGnode"] == mz)
	-sum(q_in_plus[z,t] for z in ind[:Z] if data[:Pipelines][z]["From"] == mz)
	+sum(q_out_plus[z,t] for z in ind[:Z] if data[:Pipelines][z]["To"] == mz)
	-sum(q_in_minus[z,t] for z in ind[:Z] if data[:Pipelines][z]["To"] == mz)
	+sum(q_out_minus[z,t] for z in ind[:Z] if data[:Pipelines][z]["From"] == mz)
	==sum(data[:NGDemand][d]["share"]*data[:Demand][t]["NG"] for d in ind[:NGD] if data[:NGDemand][d]["node"] == mz))

	#Solve
	@info("start solving model")
	optimize!(MILP_model)
	@info("MILP model terminates with status: $(termination_status(MILP_model))")
	if termination_status(MILP_model) == MOI.OPTIMAL
		solution = Dict(
		:obj => objective_value(MILP_model),
		:comp_time =>MOI.get(MILP_model, MOI.SolveTime()),
		:p => [value.(p[i,t]) for i in ind[:I], t in ind[:T]],
		:w => [value.(w[j,t]) for j in ind[:J], t in ind[:T]],
		:theta => [value.(theta[n,t]) for n in ind[:N], t in ind[:T]],
		:f => [value.(f[l,t]) for l in ind[:L], t in ind[:T]],
		:g => [value.(g[k,t]) for k in ind[:K], t in ind[:T]],
		:pr => [value.(pr[mz,t]) for mz in ind[:MZ], t in ind[:T]],
		:q_in_plus => [value.(q_in_plus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q_in_minus => [value.(q_in_minus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q_out_plus => [value.(q_out_plus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q_out_minus => [value.(q_out_minus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q_plus => [value.(q_plus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q_minus => [value.(q_minus[z,t]) for z in ind[:Z], t in ind[:T]],
		:q => [value.(q[z,t]) for z in ind[:Z], t in ind[:T]],
		:y => [value.(y[z,t]) for z in ind[:Z], t in ind[:T]],
		:h => linepack == true ? [value.(h[z,t]) for z in ind[:Z], t in ind[:T]] : NaN,
		:phi_m => [value.(phi_m[z,t]) for z in ind[:Z], t in ind[:T]],
		:phi_u => [value.(phi_u[z,t]) for z in ind[:Z], t in ind[:T]])
		@info("New MILP solution saved!")
		return solution
	else
		@info("No optimal solution available")
	end
end
